#Copyright 2014 MathWorks, Inc.

from matlab.engine import pythonengine

class EngineSession():
    def __init__(self):
        try:      
            pythonengine.createProcess()
            self._process_created = True
        except:
            raise
        
    def __del__(self):
        self.release()
        
    def release(self):
        if self._process_created:
            try:
                pythonengine.closeProcess()
                self._process_created = False
            except:
                pass



# Copyright 2022 MathWorks, Inc.

import os
import platform
import sys
import pkgutil

__path__ = pkgutil.extend_path(__path__, __name__)
package_folder = os.path.dirname(os.path.realpath(__file__))
sys.path.append(package_folder)

def add_dirs_to_path(bin_dir, engine_dir, extern_dir):
        """
        Adds MATLAB engine and extern/bin directories to sys.path.
        """
        path = 'PATH'

        if not os.path.isdir(engine_dir):
            raise RuntimeError("Could not find directory: {0}".format(engine_dir))
        
        if not os.path.isdir(extern_dir):
            raise RuntimeError("Could not find directory: {0}".format(extern_dir))
        
        if platform.system() == 'Windows':
            if not os.path.isdir(bin_dir):
                raise RuntimeError("Could not find directory: {0}".format(bin_dir))
            if path in os.environ:
                paths = os.environ[path]
                os.environ[path] = bin_dir + os.pathsep + paths
            else:
                os.environ[path] = bin_dir
            if sys.version_info.major >= 3 and sys.version_info.minor >= 8:
                os.add_dll_directory(bin_dir)

        sys.path.insert(0, engine_dir)
        sys.path.insert(0, extern_dir)

arch_file = os.path.join(package_folder, 'engine', '_arch.txt')
if not os.path.isfile(arch_file):
    raise RuntimeError("The MATLAB Engine for Python install is corrupted, please try to re-install.")

with open(arch_file, 'r') as root:
    [arch, bin_folder, engine_folder, extern_bin] = [line.strip() for line in root.readlines()]


add_dirs_to_path(bin_folder, engine_folder, extern_bin)

from matlabmultidimarrayforpython import double, single, uint8, int8, uint16, \
    int16, uint32, int32, uint64, int64, logical, ShapeError, SizeError



#Copyright 2017-2020 MathWorks, Inc.
import warnings
from matlab.engine import pythonengine
import sys

def _get_async_or_background_argument(kwargs):
    if 'async' in kwargs and 'background' in kwargs:
        raise KeyError(pythonengine.getMessage('EitherAsyncOrBackground'))
    background = False
    if 'async' in kwargs:
        background = kwargs.pop('async', False)
        if not isinstance(background, bool):
            raise TypeError(pythonengine.getMessage('AsyncMustBeBool'))
        if sys.version_info.major >= 3 and sys.version_info.minor >= 7:
            # No test should be passing "async" with Python 3.7 or higher, so throw an exception
            # if a test tries to do it.
            raise SyntaxError(pythonengine.getMessage('AsyncWillDeprecate'))

    if 'background' in kwargs:
        background = kwargs.pop('background', False)
        if not isinstance(background, bool):
            raise TypeError(pythonengine.getMessage('BackgroundMustBeBool'))

    if kwargs:
        raise TypeError((pythonengine.getMessage('InvalidKwargs', list(kwargs.keys())[0].__repr__())))

    return background


#Copyright 2014-2022 MathWorks, Inc.

"""
The MATLAB Engine enables you to call any MATLAB statement either synchronously
or asynchronously.  With synchronous execution, the invocation of a MATLAB
statement returns the result after the call finishes.  With asynchronous
execution, the invocation of a MATLAB statement is performed in the background 
and a FutureResult object is returned immediately.  You can call its "done" 
function to check if the call has finished, and its "result" function to obtain
 the actual result of the MATLAB statement.

This example shows how to call a MATLAB function:

>>> import matlab.engine
>>> eng = matlab.engine.start_matlab()
>>> eng.sqrt(4.0)
2.0
>>> eng.exit()
"""


import os
import sys
import importlib
import atexit
import weakref
import threading

# UPDATE_IF_PYTHON_VERSION_ADDED_OR_REMOVED : search for this string in codebase 
# when support for a Python version must be added or removed
_supported_versions = ['2_7', '3_8', '3_9', '3_10']
_ver = sys.version_info
_version = '{0}_{1}'.format(_ver[0], _ver[1])
_PYTHONVERSION = None

if _version in _supported_versions:
    _PYTHONVERSION = _version
else:
    raise EnvironmentError("Python %s is not supported." % _version)

_module_folder = os.path.dirname(os.path.realpath(__file__))
_arch_filename = os.path.join(_module_folder, "_arch.txt")
success = False 
firstExceptionMessage = ''
secondExceptionMessage = ''
try:
    pythonengine = importlib.import_module("matlabengineforpython"+_PYTHONVERSION)
except Exception as firstE:
    firstExceptionMessage = str(firstE)

if firstExceptionMessage:
    try:
        _arch_file = open(_arch_filename,'r')
        _lines = _arch_file.readlines()
        [_arch, _bin_dir,_engine_dir, _extern_bin_dir] = [x.rstrip() for x in _lines if x.rstrip() != ""]
        _arch_file.close()
        sys.path.insert(0,_engine_dir)
        sys.path.insert(0,_extern_bin_dir)

        _envs = {'win32': 'PATH', 'win64': 'PATH'}
        if _arch in _envs:
            if _envs[_arch] in os.environ:
                _env = os.environ[_envs[_arch]]
                os.environ[_envs[_arch]] = _bin_dir + os.pathsep + os.environ[_envs[_arch]]
            else:
                os.environ[_envs[_arch]] = _bin_dir
            if sys.version_info.major >= 3 and sys.version_info.minor >= 8:
                os.add_dll_directory(_bin_dir)
        pythonengine = importlib.import_module("matlabengineforpython"+_PYTHONVERSION)
    except Exception as secondE:
        str1 = 'Please reinstall MATLAB Engine for Python or contact '
        str2 = 'MathWorks Technical Support for assistance:\nFirst issue: {}\nSecond issue: {}'.format(
            firstExceptionMessage, secondE)
        secondExceptionMessage = str1 + str2

if secondExceptionMessage:        
    raise EnvironmentError(secondExceptionMessage)


"""
This lock can make sure the global variable _engines is updated correctly in
multi-thread use case.  Also, it guarantees that only one MATLAB is launched
when connect_matlab() is called if there is no shared MATLAB session.
"""
_engine_lock = threading.RLock()
_engines = []

from matlab.engine.engineerror import RejectedExecutionError
from matlab.engine.basefuture import BaseFuture
from matlab.engine.matlabfuture import MatlabFuture
from matlab.engine.fevalfuture import FevalFuture
from matlab.engine.futureresult import FutureResult
from matlab.engine.enginesession import EngineSession
from matlab.engine.matlabengine import MatlabEngine
from matlab.engine.matlabengine import enginehelper

_session = EngineSession()

def start_matlab(option="-nodesktop", **kwargs):
    """
    Start the MATLAB Engine.  This function creates an instance of the
    MatlabEngine class.  The local version of MATLAB will be launched
    with the "-nodesktop" argument.

    Please note the invocation of this function is synchronous, which
    means it only returns after MATLAB launches.
    
    Parameters
        option - MATLAB startup option.
        async, background: bool - start MATLAB asynchronously or not.  This parameter
        is optional and false by default.  "async" is a synonym for "background"
        that will be removed in a future release.
                
    Returns
        MatlabEngine - if aync or background is false.  This object can be used to evaluate
        MATLAB statements.
        FutureResult - if async or background is true.  This object can be used to obtain the
        real MatlabEngine instance.

    Raises
        EngineError - if MATLAB can't be started.
    """
    if not isinstance(option, str):
        raise TypeError(pythonengine.getMessage('StartupOptionShouldBeStr'))
    
    background = enginehelper._get_async_or_background_argument(kwargs)
    future = FutureResult(option=option)
    if not background:
        #multi-threads cannot launch MATLAB simultaneously
        eng = future.result()
        return eng
    else:
        return future

def find_matlab():
    """
    Discover all shared MATLAB sessions on the local machine. This function 
    returns the names of all shared MATLAB sessions.

    Returns
        tuple - the names of all shared MATLAB sessions running locally.
    """
    engines = pythonengine.findMATLAB()
    return engines

def connect_matlab(name=None, **kwargs):
    """
    Connect to a shared MATLAB session.  This function creates an instance 
    of the MatlabEngine class and connects it to a MATLAB session. The MATLAB 
    session must be a shared session on the local machine. 

    If name is not specified and there is no shared MATLAB available, this 
    function launches a shared MATLAB session with default options. If name 
    is not specified and there are shared MATLAB sessions available, the first 
    shared MATLAB created is connected.  If name is specified and there are no 
    shared MATLAB sessions with that name, an exception is raised. 

    Parameters 
        name: str - the name of the shared MATLAB session, which is optional.
        By default it is None.
        async, background: bool - connect to the shared MATLAB session asynchronously or
        not.  This is optional and false by default.  "async" is a synonym for 
        "background" that will be removed in a future release.

    Returns
        MatlabEngine - if async or background is false.  This object can be used to evaluate
        MATLAB functions.
        FutureResult - if async or background is true.  This object can be used to obtain the
        real MatlabEngine instance.

    Raises
        EngineError - if the MATLAB cannot be connected.
    """
    
    #multi-threads cannot run this function simultaneously 

    background = enginehelper._get_async_or_background_argument(kwargs)
    if name is None:
        with _engine_lock:
            #if there is no shareable or more than one shareable MATLAB
            engines = find_matlab()

            if len(engines) == 0:
                future = FutureResult(option="-r matlab.engine.shareEngine")
            else:
                #if there are shareable MATLAB sessions available
                future = FutureResult(name=engines[0], attach=True)

            if not background:
                eng = future.result()
                return eng
            else:
                return future
    else:
        future = FutureResult(name=name, attach=True)
        if not background:
            eng = future.result()
            return eng
        else:
            return future

@atexit.register
def __exit_engines():
    for eng in _engines:
        if eng() is not None:
            eng().exit()
    _session.release()



#Copyright 2015 MathWorks, Inc.

"""
FevalFuture: The class name of a future result returned by the MATLAB Engine.

An instance of FevalFuture is returned from each asynchronous invocation of a
MATLAB statement or function.  The future result serves as a placeholder of the
actual result, so the future result can be returned immediately.  MATLAB puts
the actual result into the placeholder when the MATLAB function finishes its
evaluation.  The future result can be used to interrupt the execution, check the
completion status, and get the result of a MATLAB statement.

"""

from matlab.engine import pythonengine
from matlab.engine import RejectedExecutionError
from matlab.engine import TimeoutError
from matlab.engine import BaseFuture
import time
import weakref

try:
    long
except NameError:
    long = int


class FevalFuture(BaseFuture):
    """
    A FevalFuture object is used to hold the future result of a MATLAB
    statement.  The FevalFuture object should be only created by MatlabEngine
    after submitting a MATLAB command for evaluation.
    """
    
    def __init__(self, eng, handle, nout, stdout, stderr):
        self._engine = weakref.ref(eng)
        self._future = handle
        self._nargout = nout
        self._out = stdout
        self._err = stderr
        self._retrieved = False
        self._result = None
        
    def result(self, timeout=None):
        """
        Get the result of a MATLAB statement.

        Parameter
            timeout: int
                    Number of seconds to wait before returning.  By default,
            this function will wait until the result is generated.

        Returns
            The result of MATLAB statement.  A tuple is returned if multiple
            outputs are returned.

        Raises
            SyntaxError - if there is an error in the MATLAB statement.
            InterruptedError - if the task is interrupted.
            CancelledError - if the evaluation of MATLAB function is cancelled already.
            TimeoutError - if this method fails to get the result in timeout seconds.
            MatlabExecutionError - if the MATLAB statement fails in execution.
            TypeError - if the data type of return value is not supported.
            RejectedExecutionError  - an error occurs if the engine is terminated.
        """ 
        self.__validate_engine()
        
        if self._retrieved:
            return self._result

        """
        Following code is used to poll the Ctrl+C every second from keyboard in
        order to cancel a MATLAB function.
        """

        try:
            result_ready = self.wait(timeout, pythonengine.waitForFEval)

            if not result_ready:
                raise TimeoutError(pythonengine.getMessage('MatlabFunctionTimeout'))

            self._result = pythonengine.getFEvalResult(self._future,self._nargout, None, out=self._out, err=self._err)
            self._retrieved = True
            return self._result

        except KeyboardInterrupt:
            self.cancel()
            if self.cancelled():
                print(pythonengine.getMessage('MatlabFunctionCancelled'))
        except:
            raise
    
    def cancel(self):
        """
        Cancel the execution of an evaluation of a MATLAB statement.
    
        Returns 
            bool - True if the corresponding MATLAB statement can be cancelled;
            False otherwise.

        Raises 
            RejectedExecutionError  - an error occurs if the engine is terminated.
        """
        self.__validate_engine()
        return pythonengine.cancelFEval(self._future)
    
    def cancelled(self):
        """
        Obtain the cancellation status of the asynchronous execution of a MATLAB
        command.
    
        Returns 
            bool - True if the execution is cancelled; False otherwise.

        Raises 
            RejectedExecutionError  - an error occurs if the engine is terminated.
        """
        self.__validate_engine()
        return pythonengine.isCancelledFEval(self._future)

    def done(self):
        """
        Obtain the completion status of the asynchronous invocation of a MATLAB
        command.

        Returns 
            bool - True if the execution is finished; False otherwise.  It
            returns True even if there is an error generated from the MATLAB
            statement or it is cancelled.

        Raises 
            RejectedExecutionError - an error occurs if the engine is terminated.
        """
        self.__validate_engine()
        return pythonengine.isDoneFEval(self._future)
    
    def __del__(self):
        if self._future is not None:
            pythonengine.destroyFEvalResult(self._future)
            self._future = None
            self._result = None
            
    def __validate_engine(self):
        if self._engine() is None or not self._engine()._check_matlab():
            raise RejectedExecutionError(pythonengine.getMessage('MatlabTerminated'))  


#Copyright 2015 MathWorks, Inc.


"""
MatlabFuture: The class name of a future handle returned by starting MATLAB
asynchronously.

An instance of MatlabFuture is returned from the invocation of starting MATLAB
asyncronously.  The future handle serves as a placeholder of the actual MATLAB
instance, the the future handle can be returned immediately.  The future handle
can be used to interrupt the launch of MATLAB, check the completion status and
get the real MATLAB Engine object.

"""

from matlab.engine import pythonengine
from matlab.engine import TimeoutError
from matlab.engine import CancelledError
from matlab.engine import BaseFuture
from matlab.engine import _engines
from matlab.engine import _engine_lock
import shlex
import weakref


class MatlabFuture(BaseFuture):
    """
    A MatlabFuture object is used to hold the future handle of a MATLAB
    instance.  The MatlabFuture object should be only created by calling
    start_matlab or connect_matlab asynchronously.
    """

    def __init__(self, **kargs):
        tokens = []
        option = kargs.pop("option", None)
        name_ = kargs.pop("name", None)
        attach = kargs.pop("attach", False)
        self._matlab = None
        self._cancelled = False
        self._done = False
        if option is not None:
            tokens = shlex.split(option)
        try:
            if attach:
                self._future = pythonengine.attachMATLABAsync(name_)
                self._attach = True
            else:
                self._future = pythonengine.createMATLABAsync(tokens)
                self._attach = False
        except:
            raise

    def result(self, timeout=None):
        """
        Get the MatlabEngine instance.

        Parameter
            timeout: Number of seconds to wait before returning.  By default,
            this function will wait until the MatlabEngine instance is ready.

        Returns
            An object of MatlabEngine class.

        Raises
            CancelledError - if the launch or connection of MATLAB is cancelled already.
            TimeoutError - if the MATLAB instance is not ready in timeout seconds.
        """
        from matlab.engine import MatlabEngine
        if self._cancelled:
            if self._attach:
                raise CancelledError(pythonengine.getMessage('ConnectMatlabCancelled'))
            else:
                raise CancelledError(pythonengine.getMessage('LaunchMatlabCancelled'))

        if self._matlab is not None:
            return self._matlab

        try:
            result_ready = self.wait(timeout, pythonengine.waitForMATLAB)

            if not result_ready:
                if self._attach:
                    raise TimeoutError(pythonengine.getMessage('ConnectMatlabTimeout'))
                else:
                    raise TimeoutError(pythonengine.getMessage('LaunchMatlabTimeout'))

            handle = pythonengine.getMATLAB(self._future)
            eng = MatlabEngine(handle)
            self._matlab = eng
            with _engine_lock:
                _engines.append(weakref.ref(eng))
            return eng

        except KeyboardInterrupt:
            self.cancel()
            print(pythonengine.getMessage('MatlabCancelled'))
        except:
            raise


    def cancel(self):
        """
        Cancel the launch of or connection to a MATLAB instance.

        Returns
            bool - True if the action can be cancelled; False otherwise.
        """
        if self._matlab is None and not self._cancelled and not self._done:
            pythonengine.cancelMATLAB(self._future)
            self._cancelled = True
            return True
        else:
            return False

    def cancelled(self):
        """
        Obtain the cancellation status of the asynchronous launch or connection
         to a MATLAB instance.

        Returns
            bool - True if the execution was cancelled; False otherwise.
        """
        if self._cancelled:
            return True
        return False

    def done(self):
        """
        Obtain the completion status of the asynchronous launch or connection
         to a MATLAB instance.

        Returns
            bool - True if the action is finished; False otherwise.
        """
        if self._done or self._cancelled or self._matlab is not None:
            return True
        self._done = pythonengine.isDoneMATLAB(self._future)
        return self._done

    def __del__(self):
        if self._future is not None:
            pythonengine.destroyMATLAB(self._future)
            self._future = None
            self._matlab = None


#Copyright 2014-2017 MathWorks, Inc.

"""
MatlabEngine: The class name of MATLAB Engine.  You can call MATLAB software as
a computational engine using the MatlabEngine class.
"""


from matlab.engine import pythonengine
from matlab.engine import FutureResult
from matlab.engine import RejectedExecutionError
from matlab.engine import MatlabExecutionError
import weakref
import shlex
from matlab.engine import enginehelper

try:
    import StringIO as sIO
except ImportError:
    import io as sIO

class MatlabFunc(object):
    """
    Reference to a MATLAB function, where "matlabfunc" is replaced by the
    function called by the user. *args are passed to MATLAB. **kwargs are
    only passed to the engine.
    """

    def __init__(self, eng, name):
        self.__dict__["_engine"] = weakref.ref(eng)
        self.__dict__["_name"] = name

    def __getattr__(self, name):
        return MatlabFunc(self._engine(), "%s.%s" % (self._name, name))

    def __setattr__(self, kw, value):
        raise AttributeError(pythonengine.getMessage('AttrCannotBeAddedToM'))

    def __call__(self, *args, **kwargs):
        self.__validate_engine()

        nargs = kwargs.pop('nargout', 1)

        if not isinstance(nargs, int):
            raise TypeError(pythonengine.getMessage('NargoutMustBeInt',  type(nargs).__name__))

        if nargs < 0:
            raise ValueError(pythonengine.getMessage('NargoutCannotBeLessThanZero'))

        _stdout = kwargs.pop('stdout', None)
        _stderr = kwargs.pop('stderr', None)

        background = enginehelper._get_async_or_background_argument(kwargs)
        
        _sIO_info = '{0}.{1}'.format(sIO.__name__, sIO.StringIO.__name__);
        if (_stdout is not None) and (not isinstance(_stdout, sIO.StringIO)):
            _stdout_info = '{0}.{1}'.format(_stdout.__class__.__module__, _stdout.__class__.__name__);
            raise TypeError(pythonengine.getMessage('StdoutMustBeStringIO', _sIO_info, _stdout_info))

        if (_stderr is not None) and (not isinstance(_stderr, sIO.StringIO)):
            _stderr_info = '{0}.{1}'.format(_stderr.__class__.__module__, _stderr.__class__.__name__);
            raise TypeError(pythonengine.getMessage('StderrMustBeStringIO', _sIO_info, _stderr_info))

        future = pythonengine.evaluateFunction(self._engine()._matlab,
                                               self._name, nargs,args,
                                               out=_stdout, err=_stderr)
        if background:
            return FutureResult(self._engine(), future, nargs, _stdout, _stderr, feval=True)
        else:
            return FutureResult(self._engine(), future, nargs, _stdout,
                                _stderr, feval=True).result()

    def __validate_engine(self):
        if self._engine() is None or not self._engine()._check_matlab():
            raise RejectedExecutionError(pythonengine.getMessage('MatlabTerminated'))  
                                
class MatlabWorkSpace(object):
    """
        ['<matlabvar>']
        ['<matlabvar>']=vardata

            Pass a variable into the MATLAB workspace and copy a
            variable from the MATLAB workspace.

            Parameters
                <matlabvar>: str
                    Variable name to be used in the MATLAB workspace.

                vardata: object
                    A Python variable to be passed into the MATLAB workspace.

            Returns
                ['<matlabvar>'] returns the variable copied from the
                MATLAB workspace.

                ['<matlabvar>']=vardata returns None.

            Raises
                NameError - if there is no such variable in the MATLAB
                workspace.

                SyntaxError - if the data is passed to the MATLAB
                workspace with an illegal variable name.

                TypeError - if <matlabvar> is not a string, or if
                the data type of vardata is not supported.

                ValueError - if <matlabvar> is empty.

                RejectedExecutionError - if the Engine is terminated.
    """

    def __init__(self,eng):
        self.__dict__["_engine"] = weakref.ref(eng)

    def __getitem__(self,attr):
        self.__validate_engine()
        self.__validate_identity(attr)
        _method=MatlabFunc(self._engine(), "matlab.internal.engine.getVariable")
        future = _method(attr)
        return future

    def __setitem__(self,attr,value):
        self.__validate_engine()
        self.__validate_identity(attr)
       
        _method=MatlabFunc(self._engine(), "assignin")
        return  _method("base", attr, value, nargout=0)
        
    def __repr__(self):
        _method = MatlabFunc(self._engine(), "whos")
        _method(nargout=0)
        return ""

    def __setattr__(self, kw, value):
        raise AttributeError(pythonengine.getMessage('AttrCannotBeAddedToMWS'))

    def __validate_engine(self):
        if self._engine() is None or not self._engine()._check_matlab():
            raise RejectedExecutionError(pythonengine.getMessage('MatlabTerminated'))

    def __validate_identity(self, attr):
        if not isinstance(attr, str):
            raise TypeError(pythonengine.getMessage('VarNameMustBeStr',  type(attr).__name__))
        if not pythonengine.validateIdentity(attr):
            raise ValueError(pythonengine.getMessage('VarNameNotValid'))
      
class MatlabEngine(object):
    """
    By default, the MATLAB Engine starts a MATLAB instance in a separate 
    process without the desktop on the local machine.  The MATLAB version 
    used by the engine application is the version of MATLAB specified in PATH.

    The MATLAB Engine supports calling MATLAB functions directly. MATLAB 
    functions are dynamically added to a MatlabEngine object as callable 
    attributes.  The function name <matlabfunc> is a replaceable MATLAB 
    function name (for example, sqrt). The function signature is the same as in
    MATLAB, with optional named arguments nargout, async|background, stdout, and stderr.

    workspace['<matlabvar>']
        A property to represent the MATLAB workspace.  Variables in
        the MATLAB workspace can be accessed through <matlabvar>.  The
        type of this property is MatlabWorkSpace.


    <matlabfunc>(*args, nargout=1, async=False, stdout=sys.stdout, stderr=sys.stderr)
    <matlabfunc>(*args, nargout=1, background=False, stdout=sys.stdout, stderr=sys.stderr)

        The invocation of a MATLAB statement can be either synchronous
        or asynchronous.  While a synchronous function call returns
        the result after it finishes executing, an asynchronous
        function call is performed in the background and returns a FutureResult
        immediately.  This FutureResult object can be used to retrieve the 
        actual result later.  If there are any output or error messages 
        generated from <matlabfunc>, by default they will be redirected to the
        standard output or standard error of the Python console.

        Please note that you can call an arbitrary MATLAB function
        available in the MATLAB path using feval and eval.

        Parameters
            args:
                Arguments accepted by the MATLAB function to be called.
            nargout: int
                By default, the number of output is 1.  If the number of output
            is more than 1, a tuple is returned.
            async, background: bool
                This parameter is used to specify how the MATLAB command is
            evaluated: asynchronously or synchronously. By default, async|background
            is chosen to be False so the MATLAB command is evaluated synchronously.
            "async" is a synonym for "background" that will be removed in a future release.
            stdout: StringIO.StringIO (Python 2.7),  io.StringIO (Python 3)
                Stream used to capture the output of the MATLAB command.  By
            default, the system standard output sys.stdout is used.
            stderr: StringIO.StringIO (Python 2.7),  io.StringIO (Python 3)
                Stream used to capture the error message of the MATLAB command.
            By default, the system standard error output sys.stderr is used.

        Returns
            The type of the return value of this function varies based on the
            value of parameter async or background.  For the case of synchronously invocation,
            the result of the MATLAB command is returned directly.  For the case
            of asynchronous invocation, a FutureResult is returned which can be
            used to retrieve the actual result, check completion status, and
            interrupt the execution of the MATLAB function.

        Raises
            RejectedExecutionError - if the engine is terminated.
            SyntaxError - if there is an error in parsing the MATLAB statement.
            MatlabExecutionError - if the MATLAB statement fails in execution.
            TypeError - if the data types of *args are not supported by
            MATLABEngine; or if the data type of a return value is not supported.

    """

    def __init__(self, matlab):
        self.__dict__["_matlab"] = matlab
        self.__dict__["workspace"] = MatlabWorkSpace(self)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.exit()

    def exit(self):
        """
        Stop the MATLAB session.  Calling this method will terminate the
        MatlabEngine instance immediately.
        """
        if self._check_matlab():
            pythonengine.closeMATLAB(self.__dict__["_matlab"])
            self.__dict__.pop("_matlab")

    def quit(self):
        """
        Stop the MATLAB session.  Calling this method will terminate the
        MatlabEngine instance immediately.
        """
        self.exit()

    def __getattr__(self,name):
        """Dynamic attribute of MatlabEngine"""
        return MatlabFunc(self, name)
    
    def __setattr__(self, kw, value):
        raise AttributeError(pythonengine.getMessage('AttrCannotBeAddedToM'))

    def __del__(self):
        self.exit()
        
    def _check_matlab(self):
         return "_matlab" in self.__dict__



#Copyright 2014 MathWorks, Inc.

class RejectedExecutionError(Exception):
    """Exception raised from MATLAB engine"""
    
    def __init__(self,  message):
        self.message = message
    def __repr__(self):
        return self.message
        
        
    


#Copyright 2014-2017 MathWorks, Inc.

"""
FutureResult: The class name of a future result returned by the MATLAB Engine.

An instance of FutureResult is returned from each asynchronous invocation of a
function call: start_matlab, connect_matlab, or MatlabEngine.<matlabfunc>.  The
 future result serves as a placeholder of the actual result, so the future
 result can be returned immediately.  The actual result is placed into the
 placeholder when the function finishes its evaluation.  The future result can
 be used to interrupt the execution, check the completion status, and get the
 actual result.
"""

from matlab.engine import pythonengine
from matlab.engine import MatlabFuture
from matlab.engine import FevalFuture

try:
    long
except NameError:
    long = int

class FutureResult():
    """
    A FutureResult object is used to hold the future result of a function call.
    The FutureResult object can be created by start_matlab(async=True),
    connect_matlab(async=True), and MaltabEngine.<matlabfunc>(async=True).
    """
    
    def __init__(self, *args, **kwargs):
        feval = kwargs.pop("feval", False)
        if feval:
            self.__future = FevalFuture(*args)
        else:
            self.__future = MatlabFuture(*args, **kwargs)

    def result(self, timeout=None):
        """
        Get the MatlabEngine object or the result of a MATLAB statement.

        Parameter
            timeout: int
                    Number of seconds to wait before returning.  By default,
            this function will wait until the result is generated.

        Returns
            The MatlabEngine object or the result of MATLAB statement.  A tuple
             is returned if multiple outputs are returned.

        Raises
            SyntaxError - if there is an error in the MATLAB statement.
            InterruptedError - if the task is interrupted.
            CancelledError - if the evaluation of MATLAB function is cancelled already.
            TimeoutError - if this method fails to get the result in timeout seconds.
            MatlabExecutionError - if the MATLAB statement fails in execution.
            TypeError - if the data type of return value is not supported.
            RejectedExecutionError  - an error occurs if the engine is terminated.
        """
        if timeout is not None:
            if not isinstance(timeout, (int, long, float)):
                raise TypeError(pythonengine.getMessage('TimeoutMustBeNumeric', type(timeout).__name__))
            
            if timeout < 0:
                raise TypeError(pythonengine.getMessage('TimeoutCannotBeNegative'))
        
        return self.__future.result(timeout)
    
    def cancel(self):
        """
        Cancel the launch/connection of MATLAB or evaluation of a MATLAB
        task.
    
        Returns 
            bool - True if the corresponding task can be cancelled;
            False otherwise.

        Raises 
            RejectedExecutionError  - an error occurs if the Engine is terminated.
        """
        return self.__future.cancel()
    
    def cancelled(self):
        """
        Obtain the cancellation status of the asynchronous execution of the
        function.
    
        Returns 
            bool - True if the execution is cancelled; False otherwise.

        Raises 
            RejectedExecutionError  - an error occurs if the engine is terminated.
        """
        return self.__future.cancelled()

    def done(self):
        """
        Obtain the completion status of the asynchronous invocation of the
        task.

        Returns 
            bool - True if the execution is finished; False otherwise.  It
            returns True even if there is an error generated from the task
            or the task is cancelled.

        Raises 
            RejectedExecutionError - an error occurs if the engine is terminated.
        """
        return self.__future.done()
    
    def __del__(self):
        if self.__future is not None:
            del self.__future


#Copyright 2015 MathWorks, Inc.

"""
BaseFuture: The base class of FevalFuture and MatlabFuture.

"""

import time


class BaseFuture(object):

    def wait(self, timeout, wait_for_func):
        """
        Wait for the execution of a function.

        Parameter
            timeout: int
                    Number of seconds to wait before returning.

        Returns
            The result is ready or not.
        """
        time_slice = 1
        if timeout is None:
            result_ready = self.done()
            while not result_ready:
                result_ready = wait_for_func(self._future, time_slice)
        else:
            result_ready = self.done()
            current_time = time.time()
            sleep_until = current_time + timeout
            while (not result_ready) and (current_time < sleep_until):
                if (sleep_until - current_time) >= time_slice:
                    result_ready = wait_for_func(self._future, time_slice)
                else:
                    time_to_sleep = sleep_until - current_time
                    result_ready = wait_for_func(self._future, time_to_sleep)
                current_time = time.time()
        return result_ready


# Copyright 2022-2023 Mathworks, Inc.

from setuptools import setup, find_packages
from setuptools.command.build_py import build_py 
import os
import re
import sys
import platform
import xml.etree.ElementTree as xml
if platform.system() == 'Windows':
    import winreg

class _MatlabFinder(build_py):
    """
    Private class that finds MATLAB on user's computer prior to package installation.
    """
    PLATFORM_DICT = {
        'Windows': 'PATH', 
        'Linux': 'LD_LIBRARY_PATH', 
        'Darwin': 'DYLD_LIBRARY_PATH'
    }
    
    # MUST_BE_UPDATED_EACH_RELEASE (Search repo for this string)
    MATLAB_REL = 'R2022b'

    # MUST_BE_UPDATED_EACH_RELEASE (Search repo for this string)
    MATLAB_VER = '9.13.7'

    # MUST_BE_UPDATED_EACH_RELEASE (Search repo for this string)
    SUPPORTED_PYTHON_VERSIONS = set(['3.8', '3.9', '3.10'])

    # MUST_BE_UPDATED_EACH_RELEASE (Search repo for this string)
    VER_TO_REL = {
        "9.9": "R2020b",
        "9.10": "R2021a",
        "9.11": "R2021b",
        "9.12": "R2022a",
        "9.13": "R2022b"
    }

    DEFAULT_INSTALLS = {
        'Darwin': f"/Applications/MATLAB_{MATLAB_REL}.app",
        'Linux': f"/usr/local/MATLAB/{MATLAB_REL}"
    }

    arch = ''
    path_env_var_name = ''
    python_ver = ''
    platform = ''
    found_matlab_release = ''
    found_matlab_version = ''
    found_matlab_with_wrong_arch_in_default_install = ''
    found_matlab_with_wrong_arch_in_path = ''
    verbose = False
    
    # ERROR MESSAGES
    minimum_maximum = "No compatible version of MATLAB was found. " + \
        "Version {this_v:s} was found, but this feature only supports MATLAB {min_v:s} ({min_r:s}) through {max_v:s} ({max_r:s}), inclusive."
    dir_not_found = "Directory not found: "
    no_windows_install = "MATLAB installation not found in Windows Registry:"
    unsupported_platform = "{platform:s} is not a supported platform."
    unsupported_python = "{python:s} is not supported. The supported Python versions are {supported:s}."
    unset_env = "Environment variable {path1:s} has not been set. Add <matlabroot>/bin/{arch:s} to {path2:s}, where <matlabroot> is the root of a valid MATLAB installation."
    install_or_set_path = "MATLAB {ver:s} installation not found. Install to default location, or add <matlabroot>/bin/{arch:s} to {path:s}, where <matlabroot> is the root of a MATLAB {ver:s} installation."
    no_compatible_matlab = "No compatible MATLAB installation found in Windows Registry. This release of " + \
        "MATLAB Engine API for Python is compatible with version {ver:s}. The found versions were"
    no_matlab = "No compatible MATLAB installation found in Windows Registry."
    incompatible_ver = "MATLAB version {ver:s} ({rel:s}) was found, but this release of MATLAB Engine API for Python is not compatible with it. " + \
        "To install a compatible version, call 'python -m pip install matlabengine=={ver:s}'."
    invalid_version_from_matlab_ver = "Format of MATLAB version '{ver:s}' is invalid."
    invalid_version_from_eng = "Format of MATLAB Engine API version '{ver:s}' is invalid."
    next_steps = "Reinstall MATLAB, use DYLD_LIBRARY_PATH to specify a different MATLAB installation, or use a different Python interpreter."
    wrong_arch_in_default_install = "MATLAB installation in {path1:s} is {matlab_arch:s}, but Python interpreter is {python_arch:s}. {next_steps:s}."
    wrong_arch_in_path = "MATLAB installation in {path1:s}, listed in DYLD_LIBRARY_PATH, is {matlab_arch:s}, but Python interpreter is {python_arch:s}. {next_steps:s}."
    
    def _print_if_verbose(self, msg):
        if self.verbose:
            print(msg)
            
    def set_platform_and_arch(self):
        """
        Sets the platform and architecture. 
        """
        self.platform = platform.system()
        if self.platform not in self.PLATFORM_DICT:
            raise RuntimeError(self.unsupported_platform.format(platform=self.platform))
        else:
            self.path_env_var_name = self.PLATFORM_DICT[self.platform]
        
        if self.platform == 'Windows':
            self.arch = 'win64'
        elif self.platform == 'Linux':
            self.arch = 'glnxa64'
        elif self.platform == 'Darwin':
            if platform.mac_ver()[-1] == 'arm64':
                self.arch = 'maca64'
            else:
                self.arch = 'maci64'
        else:
            raise RuntimeError(self.unsupported_platform.format(platform=self.platform))
    
    def set_python_version(self):
        """
        Gets Python version and ensures it is supported.
        """
        ver = sys.version_info
        self.python_ver = f"{ver.major}.{ver.minor}"

        if self.python_ver not in self.SUPPORTED_PYTHON_VERSIONS:
            raise RuntimeError(self.unsupported_python.format(python=self.python_ver, supported=str(self.SUPPORTED_PYTHON_VERSIONS)))

    def unix_default_install_exists(self):
        """
        Determines whether MATLAB is installed in default UNIX location.
        """
        path = self.DEFAULT_INSTALLS[self.platform]
        if not os.path.exists(path):
            return False
        
        if self.platform == 'Darwin':
            # On Mac, we need to further verify that there is a 'bin/maci64' subdir if the Python is maci64
            # or a 'bin/maca64' subdir if the Python is maca64.
            path_to_bin = os.path.join(path, 'bin', self.arch)
            if os.path.exists(path_to_bin):
                # The path exists, and we don't need to do anything further.
                return True
                
            if self.arch == 'maci64':
                alternate_arch = 'maca64'
            else:
                alternate_arch = 'maci64'
                
            if os.path.exists(os.path.join(path, 'bin', alternate_arch)):
                # There is a default install, but its arch doesn't match the Python arch. Save this info
                # so that if we don't find an install with a valid arch in DYLD_LIBRARY_PATH, we can
                # issue an error message that says that there is a Mac installation in the default 
                # location that has the wrong arch. The user can choose whether to change the
                # Python interpreter or the MATLAB installation so that the arch will match.
                self.found_matlab_with_wrong_arch_in_default_install = path
                return False
                
        return True
    
    def _create_path_list(self):
        """
        Creates a list of directories on the path to be searched.
        """
        path_dirs = []
        path_string = ''
        if self.path_env_var_name in os.environ:
            path_string = os.environ[self.path_env_var_name]
            path_dirs.extend(path_string.split(os.pathsep))
        
        if not path_dirs:
            raise RuntimeError(self.install_or_set_path.format(
                ver=self.MATLAB_REL, arch=self.arch, 
                    path=self.path_env_var_name))
                    
        self._print_if_verbose(f'_create_path_list returned: {path_dirs}')
        return path_dirs
    
    def _get_alternate_arch(self):
        if self.arch == 'maci64':
            return 'maca64'
        if self.arch == 'maca64':
            return 'maci64'
        return self.arch

    def _arch_in_mac_dir_is_correct(self, dir):
        ARCH_LEN = 6 # == len('maci64') or len('maca64')
        BIN_ARCH_LEN = ARCH_LEN + 4 # == len('bin/maci64') or len('bin/maca64')
        
        if len(dir) < BIN_ARCH_LEN:
            return False
        
        if dir[-1] == os.sep:
            # It's safe to look at dir[[-1 * (ARCH_LEN+1)] because BIN_ARCH_LEN > ARCH_LEN + 1.
            possible_arch = dir[-1 * (ARCH_LEN+1) : -1]
        else:
            possible_arch = dir[-1 * ARCH_LEN :]
        
        self._print_if_verbose(f'possible_arch: {possible_arch}; self.arch: {self.arch}')
        if possible_arch == self.arch:
            return True
        else:
            return False            
            
    def _get_matlab_root_from_unix_bin(self, dir):
        """
        Searches bin directory for presence of MATLAB file. Used only for
        UNIX systems. 
        """
        matlab_path = os.path.join(dir, 'MATLAB')
        possible_root = os.path.normpath(os.path.join(dir, os.pardir, os.pardir))
        matlab_root = ''
        if os.path.isfile(matlab_path) and self.verify_matlab_release(possible_root):
            if self.platform == 'Darwin' and not self._arch_in_mac_dir_is_correct(dir):
                self.found_matlab_with_wrong_arch_in_path = possible_root
                self._print_if_verbose(f'self.found_matlab_with_wrong_arch_in_path: {self.found_matlab_with_wrong_arch_in_path}')
            else:
                matlab_root = possible_root
        self._print_if_verbose(f'_get_matlab_root_from_unix_bin returned: {matlab_root}')
        return matlab_root
    
    def get_matlab_root_from_windows_reg(self):
        """
        Searches Windows Registry for MATLAB installs and gets the root directory of MATLAB.
        """
        try:
            reg = winreg.ConnectRegistry(None, winreg.HKEY_LOCAL_MACHINE)
            key = winreg.OpenKey(reg, "SOFTWARE\\MathWorks\\MATLAB")
        except OSError as err:
            raise RuntimeError(f"{self.no_windows_install} {err}")
        
        matlab_ver_key = self._find_matlab_key_from_windows_registry(key)
        ret = self._get_root_from_version_key(reg, matlab_ver_key)
        self._print_if_verbose(f'get_matlab_root_from_windows_reg returned: {ret}')
        return ret
    
    def _get_root_from_version_key(self, reg, ver_key):
        """
        Opens a registry corresponding to the version of MATLAB specified and queries for
        MATLABROOT.
        """
        try:
            key = winreg.OpenKey(reg, "SOFTWARE\\MathWorks\\MATLAB\\" + ver_key)
            matlab_root = winreg.QueryValueEx(key, "MATLABROOT")[0]
        except (OSError, FileNotFoundError) as err:
            raise RuntimeError(f"{self.no_windows_install} {err}")
        
        self._print_if_verbose(f'_get_root_from_version_key returned: {matlab_root}')
        return matlab_root
    
    def _find_matlab_key_from_windows_registry(self, key):
        """
        Searches the MATLAB folder in Windows registry for the specified version of MATLAB. When found, 
        the MATLAB root directory will be returned.
        """
        # QueryInfoKey returns a tuple, index 0 is the number of sub keys we need to search
        num_keys = winreg.QueryInfoKey(key)[0]
        key_value = ''
        found_vers = []
        for idx in range(num_keys):
            sub_key = winreg.EnumKey(key, idx)
            if sub_key in self.VER_TO_REL:
                found_vers.append(sub_key)
                # Example: the version in the registry could be "9.X" whereas the version in this file could be "9.X.Y".
                # We want to allow this.
                if self._check_matlab_ver_against_engine(sub_key):
                    key_value = sub_key
                    break
        
        if not key_value:
            if found_vers:
                vers = ', '.join(found_vers)
                eng_ver_major_minor = self._get_engine_ver_major_minor(self.MATLAB_VER)
                eng_ver_major_minor_as_str = '{}.{}'.format(eng_ver_major_minor[0], eng_ver_major_minor[1])
                raise RuntimeError(f"{self.no_compatible_matlab.format(ver=eng_ver_major_minor_as_str)} {vers}.")
            else:
                raise RuntimeError(f"{self.no_matlab}")

        self._print_if_verbose(f'_find_matlab_key_from_windows_registry returned: {key_value}')
        return key_value       

    def _get_engine_ver_major_minor(self, id):
        re_major_minor = "^(\d+)\.(\d+)"
        eng_match = re.match(re_major_minor, id)
        if not eng_match:
            raise RuntimeError(f"{self.invalid_version_from_eng.format(ver=self.MATLAB_VER)}")
        ret = (eng_match.group(1), eng_match.group(2))
        self._print_if_verbose(f'_get_engine_ver_major_minor returned: {ret}')
        return ret
        
    def _check_matlab_ver_against_engine(self, matlab_ver):
        re_major_minor = "^(\d+)\.(\d+)"
        matlab_ver_match = re.match(re_major_minor, matlab_ver)
        if not matlab_ver_match:
            raise RuntimeError(f"{self.invalid_version_from_matlab_ver.format(ver=matlab_ver)}")
        eng_major_minor = self._get_engine_ver_major_minor(self.MATLAB_VER)
        matlab_ver_major_minor = (matlab_ver_match.group(1), matlab_ver_match.group(2))
        return (matlab_ver_major_minor == eng_major_minor)
    
    def verify_matlab_release(self, root):
        """
        Parses VersionInfo.xml to verify that the MATLAB release matches the supported release
        for the Python Engine.
        """
        version_info = os.path.join(root, 'VersionInfo.xml')
        if not os.path.isfile(version_info):
            return False
        
        tree = xml.parse(version_info)
        tree_root = tree.getroot()

        matlab_release = ''
        for child in tree_root:
            if child.tag == 'release':
                matlab_release = self.found_matlab_release = child.text
            elif child.tag == 'version':
                major, minor = self._get_engine_ver_major_minor(child.text)
                self.found_matlab_version = f'{major}.{minor}'
        return matlab_release == self.MATLAB_REL

    def search_path_for_directory_unix(self, arch, path_dirs):
        """
        Used for finding MATLAB root in UNIX systems. Searches all paths ending in
        /bin/<arch> for the presence of MATLAB file to ensure the path is within
        the MATLAB tree. 
        """
        dir_to_find = os.path.join('bin', arch)
        # directory could end with slashes
        endings = [dir_to_find, dir_to_find + os.sep]

        matlab_root = ''
        dir_idx = 0
        while not matlab_root and dir_idx < len(path_dirs):
            path = path_dirs[dir_idx]
            ending_idx = 0
            while not matlab_root and ending_idx < len(endings):
                ending = endings[ending_idx]
                if path.endswith(ending):
                    # _get_matlab_root_from_unix_bin will return an empty string if MATLAB is not found.
                    # Non-empty string (MATLAB found) will break both loops.
                    matlab_root = self._get_matlab_root_from_unix_bin(path)
                ending_idx += 1
            dir_idx += 1
        self._print_if_verbose(f'search_path_for_directory_unix returned: {matlab_root}')
        return matlab_root
    
    def _err_msg_if_bad_matlab_root(self, matlab_root):
        if not matlab_root:
            if self.found_matlab_version:
                self._print_if_verbose(f'self.found_matlab_version: {self.found_matlab_version}; self.VER_TO_REL: {self.VER_TO_REL}')
                if self.found_matlab_version in self.VER_TO_REL:
                    return self.incompatible_ver.format(ver=self.found_matlab_version, rel=self.found_matlab_release)
                # Found a MATLAB release but it is older than the oldest version supported,
                # or newer than the newest version supported.
                else:
                    v_to_r_keys = list(self.VER_TO_REL.keys())
                    self._print_if_verbose(f'v_to_r_keys: {v_to_r_keys}')
                    min_v = v_to_r_keys[0]
                    min_r = self.VER_TO_REL[min_v]
                    max_v = v_to_r_keys[-1]
                    max_r = self.VER_TO_REL[max_v]
                    return self.minimum_maximum.format(this_v=self.found_matlab_release, min_v=min_v, min_r=min_r, max_v=max_v, max_r=max_r)
            else:
                # If we reach this line, we assume that the default location has already been checked for an
                # appropriate MATLAB installation but none was found.
                return self.install_or_set_path.format(ver=self.MATLAB_REL, arch=self.arch, 
                    path=self.path_env_var_name)
        
        if not os.path.isdir(matlab_root):
            return f"{self.dir_not_found} {matlab_root}"
            
        return ''
        
    def write_text_file(self, matlab_root):
        """
        Writes root.txt for use at import time.
        """
        file_location = os.path.join(os.getcwd(), 'src', 'matlab', 'engine', '_arch.txt')
        bin_arch = os.path.join(matlab_root, 'bin', self.arch)
        engine_arch = os.path.join(matlab_root, 'extern', 'engines', 'python', 'dist', 'matlab', 'engine', self.arch)
        extern_bin = os.path.join(matlab_root, 'extern', 'bin', self.arch)
        with open(file_location, 'w') as root_file:
            root_file.write(self.arch + '\n')
            root_file.write(bin_arch + '\n')
            root_file.write(engine_arch + '\n')
            root_file.write(extern_bin)

    def run(self):
        """
        Logic that runs prior to installation.
        """
        self.set_platform_and_arch()
        self.set_python_version()

        if self.platform == 'Windows':
            matlab_root = self.get_matlab_root_from_windows_reg()
        else:
            if self.unix_default_install_exists():
                matlab_root = self.DEFAULT_INSTALLS[self.platform]
            else:
                path_dirs = self._create_path_list()
                matlab_root = self.search_path_for_directory_unix(self.arch, path_dirs)
            err_msg = self._err_msg_if_bad_matlab_root(matlab_root)
            if err_msg:
                if self.platform == 'Darwin':
                    if self.found_matlab_with_wrong_arch_in_default_install:
                        raise RuntimeError(
                            self.wrong_arch_in_default_install.format(
                                path1=self.found_matlab_with_wrong_arch_in_default_install,
                                matlab_arch=self._get_alternate_arch(),
                                python_arch=self.arch,
                                next_steps=self.next_steps))
                    if self.found_matlab_with_wrong_arch_in_path:
                        raise RuntimeError(
                            self.wrong_arch_in_path.format(
                                path1=self.found_matlab_with_wrong_arch_in_path,
                                matlab_arch=self._get_alternate_arch(),
                                python_arch=self.arch,
                                next_steps=self.next_steps))
                raise RuntimeError(err_msg)

        self.write_text_file(matlab_root)
        build_py.run(self)


if __name__ == '__main__':
    with open('README.md', 'r', encoding='utf-8') as rm:
        long_description = rm.read()

    setup(
        name="matlabengine",
        # MUST_BE_UPDATED_EACH_RELEASE (Search repo for this string)
        version="9.13.7",
        description='A module to call MATLAB from Python',
        author='MathWorks',
        license="MathWorks XSLA License",
        url='https://github.com/mathworks/matlab-engine-for-python/',
        long_description=long_description,
        long_description_content_type="text/markdown",
        package_dir={'': 'src'},
        packages=find_packages(where="src"),
        cmdclass={'build_py': _MatlabFinder},
        package_data={'': ['_arch.txt']},
        zip_safe=False,
        project_urls={
            'Documentation': 'https://www.mathworks.com/help/matlab/matlab-engine-for-python.html',
            'Source': 'https://github.com/mathworks/matlab-engine-for-python',
            'Tracker': 'https://github.com/mathworks/matlab-engine-for-python/issues'
        },
        keywords=[
            "MATLAB",
            "MATLAB Engine for Python"
        ],
        classifiers=[
            "Natural Language :: English",
            "Intended Audience :: Developers",
            # MUST_BE_UPDATED_EACH_RELEASE (Search repo for this string)
            "Programming Language :: Python :: 3.8",
            "Programming Language :: Python :: 3.9",
            "Programming Language :: Python :: 3.10"
        ],
        # MUST_BE_UPDATED_EACH_RELEASE (Search repo for this string)
        python_requires=">=3.8, <3.11"
    )
